#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>

#include "porting_defines.h"
#include "crisSisGetUtilities.h"
#include "utilities.h"

    /* sort routine and struct name used therein */
static char *structNameForSortRoutine = NULL;
static int sortByTimeStamp(const void *pStruct1, const void *pStruct2);


	/* this routine allocates memory and points pStructs at it, and
	 * fills it in with overlap structures.  this memory must be
	 * deallocated outside of this routine.
	 */
void *getOverlapStructs(uint32 secondsSince1996Start,
				uint32 secondsSince1996Stop, char *structName) {
    int *pDaysOverlapStructCount_0;
    int *pDaysOverlapStructCount;
    void *pTempOverlapStructs_0, *pOverlapStructs_0;
    void *pTempOverlapStructs, *pOverlapStructs;
    char overlapFileName[256], dataFileSuffix[5], instrumentID;
    int i, structSize, period, seconds, zerothDay, lastDay;
    int overlapStructTotal, day, goodOverlapStructCount;

    period = getDataTypePeriodFromStructName(structName);
    if (period > 0) {
	seconds = secondsSince1996Start - period + 1;
	zerothDay = getDay1996FromSecondsSince1996(seconds);
    }
    else {
	zerothDay = getDay1996FromSecondsSince1996(secondsSince1996Start) - 1;
    }
    lastDay = getDay1996FromSecondsSince1996(secondsSince1996Stop - 1);

    pDaysOverlapStructCount_0 = calloc((lastDay - zerothDay + 1), sizeof(int));
    overlapStructTotal = 0;
    pDaysOverlapStructCount = pDaysOverlapStructCount_0;
    for (day = zerothDay; day <= lastDay; day++) {
	*pDaysOverlapStructCount = getOverlapStructCount(day, structName);
	overlapStructTotal += *pDaysOverlapStructCount;
	pDaysOverlapStructCount++;
    }

    structSize = getStructSizeFromName(structName);

    if (overlapStructTotal == 0) {
	free(pDaysOverlapStructCount_0);
	pOverlapStructs = calloc(1, structSize);
	memset(pOverlapStructs, '\0', structSize);
	return pOverlapStructs;
    }

    instrumentID = getInstrumentIDFromStructName(structName);
    getDataFileSuffixFromStructName(dataFileSuffix, structName);

    pTempOverlapStructs_0 = calloc(overlapStructTotal, structSize);
    pTempOverlapStructs = pTempOverlapStructs_0;
    pDaysOverlapStructCount = pDaysOverlapStructCount_0;
    for (day = zerothDay; day <= lastDay; day++) {
	int fd, bytesToRead;
	struct tm *pTimeStruct;

	if (*pDaysOverlapStructCount == 0) {
	    pDaysOverlapStructCount++;
	    continue;
	}

	pTimeStruct = getDateFromDay1996(day);
	generateDataFilePath(overlapFileName, pTimeStruct,
                                        instrumentID, dataFileSuffix);
	strcat(overlapFileName, ".overlap");

	fd = open(overlapFileName, O_RDONLY | O_BINARY, 0);
	bytesToRead = *pDaysOverlapStructCount * structSize;
	readFromFile(fd, (char *) pTempOverlapStructs, bytesToRead);
	close(fd);
	pDaysOverlapStructCount++;
	    /* very rigorous pointer addition for Windows compiler */
	pTempOverlapStructs =
		(void *) (((char *) pTempOverlapStructs) + bytesToRead);
    }

    free(pDaysOverlapStructCount_0);

    pOverlapStructs_0 = calloc(overlapStructTotal + 1, structSize);
    pOverlapStructs = pOverlapStructs_0;
    pTempOverlapStructs = pTempOverlapStructs_0;
    goodOverlapStructCount = 0;
    for (i = 0; i < overlapStructTotal; i++) {
	uint32 lastValidSecond;

	lastValidSecond =
		getLastValidSecondOfStruct(pTempOverlapStructs, structName);
	if ((lastValidSecond >= secondsSince1996Start) &&
				(lastValidSecond < secondsSince1996Stop)) {
	    memcpy(pOverlapStructs, pTempOverlapStructs, structSize);
		/* very rigorous pointer addition for Windows compiler */
	    pOverlapStructs = (void *) (((char *)pOverlapStructs) + structSize);
	    goodOverlapStructCount++;
	}
	    /* very rigorous pointer addition for Windows compiler */
	pTempOverlapStructs =
		(void *) (((char *) pTempOverlapStructs) + structSize);
    }
    free(pTempOverlapStructs_0);

    pOverlapStructs_0 = realloc(pOverlapStructs_0,
				structSize * (goodOverlapStructCount + 1));
    structNameForSortRoutine = structName;
    qsort(pOverlapStructs_0, goodOverlapStructCount,
					structSize, sortByTimeStamp);
	/* very rigorous pointer addition for Windows compiler */
    pOverlapStructs = (void *)
	(((char *) pOverlapStructs_0) + (goodOverlapStructCount * structSize));
    memset(pOverlapStructs, '\0', structSize);
    pOverlapStructs = pOverlapStructs_0;

    return pOverlapStructs;
}

static int sortByTimeStamp(const void *pStruct1, const void *pStruct2) {
    char *structName;
    uint32 timeStamp1, timeStamp2;
 
    structName = structNameForSortRoutine;
    timeStamp1 = getTimeStamp(pStruct1, structName);
    timeStamp2 = getTimeStamp(pStruct2, structName);
 
    if (timeStamp1 > timeStamp2) { return (1); }
    else if (timeStamp1 < timeStamp2) { return (-1); }
    else { return (0); }
}
